/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.service.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.springframework.stereotype.Service;

import kieker.webgui.common.exception.GraphLayoutException;
import kieker.webgui.service.IGraphLayoutService;

import org.eclipse.emf.common.util.EList;

import de.cau.cs.kieler.core.alg.BasicProgressMonitor;
import de.cau.cs.kieler.core.alg.IKielerProgressMonitor;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.AbstractLayoutProvider;
import de.cau.cs.kieler.kiml.klayoutdata.KEdgeLayout;
import de.cau.cs.kieler.kiml.klayoutdata.KPoint;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.EdgeRouting;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.PortConstraints;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klay.layered.LayeredLayoutProvider;

/**
 * This is a utility class providing a method to layout a graph from the graph flow editor. The method accepts two strings (one for the nodes, one for the edges)
 * which are the result of specific javascript functions of the graph. For further information look into the API description of the flow editor itself.
 * 
 * @author row, Nils Christian Ehmke
 */
@Service
public final class GraphLayoutServiceImpl implements IGraphLayoutService {

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public GraphLayoutServiceImpl() {
		// No code necessary
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see kieker.webgui.service.impl.IGraphLayoutService#layoutGraph(java.lang.String, java.lang.String)
	 */
	@Override
	public String layoutGraph(final String nodes, final String edges) {
		try {
			final LayoutInformation layoutInformation = GraphLayoutServiceImpl.assembleLayoutInformation(nodes, edges);

			// Create a progress monitor
			final IKielerProgressMonitor progressMonitor = new BasicProgressMonitor();

			// Create the layout provider
			final AbstractLayoutProvider layoutProvider = new LayeredLayoutProvider();

			// Perform layout on the created graph
			layoutProvider.doLayout(layoutInformation.getGraph(), progressMonitor);

			// convert layouted graph positions to String
			final String layoutedGraph = GraphLayoutServiceImpl.getPositions(layoutInformation);
			final String bendPoints = GraphLayoutServiceImpl.getBendPoints(layoutInformation);

			final StringBuilder sb = new StringBuilder("autoLayout#");
			sb.append(layoutedGraph);
			sb.append("#");
			sb.append(bendPoints);

			return sb.toString();
		} catch (final GraphLayoutException e) {
			e.printStackTrace();
			return "Error#Empty Graph!";
		} catch (final NumberFormatException e) {
			e.printStackTrace();
			return "Error#Invalid Graph Information String!";
		} catch (final Exception e) { // NOCS
			e.printStackTrace();
			return "Error#Unknown Error!";
		}
	}

	/**
	 * Assembles the necessary layout information using the given parameters.
	 * 
	 * @param nodesStr
	 *            The string containing the information about the nodes.
	 * @param edgesStr
	 *            The string containing the information about the edges.
	 * 
	 * @return An object containing the layout information for the graph.
	 */
	private static LayoutInformation assembleLayoutInformation(final String nodesStr, final String edgesStr) throws NumberFormatException, GraphLayoutException {
		final int firstSemicolon = nodesStr.indexOf(';');

		// if there is no semicolon, we got an empty graph
		if (firstSemicolon < 0) {
			throw new GraphLayoutException();
		}

		final int scale = Integer.parseInt(nodesStr.substring(0, firstSemicolon));

		final String[] nodeInfo = GraphLayoutServiceImpl.correctEmptyArray(nodesStr.substring(firstSemicolon + 1).split(";"));
		final String[] edgeInfo = GraphLayoutServiceImpl.correctEmptyArray(edgesStr.split(";"));

		final Map<String, List<Object>> children = new HashMap<String, List<Object>>();
		final Object[][] edges = new Object[edgeInfo.length][5];
		final LayoutInformation layoutInformation = new LayoutInformation(KimlUtil.createInitializedNode(), children, edges);

		GraphLayoutServiceImpl.addNodes(scale, nodeInfo, layoutInformation);
		GraphLayoutServiceImpl.addEdges(edgeInfo, layoutInformation);
		return layoutInformation;
	}

	/**
	 * Parses a String array, building a graph with nodes and fixed ports.
	 * 
	 * @param scale
	 *            The scale factor of the graph.
	 * @param nodeInfo
	 *            Information about the nodes: isItCustom?, width, height, parent, portInformation
	 * @param layoutInformation
	 *            The necessary information about the graph.
	 * 
	 * @throws NumberFormatException
	 *             The nodeInfo should mostly be space separated numbers. If it is however invalid, this
	 *             Exception will be thrown
	 */
	private static void addNodes(final int scale, final String[] nodeInfo, final LayoutInformation layoutInformation) throws NumberFormatException {
		// define node HashMap port list
		final Map<String, List<Object>> nodesAndPorts = layoutInformation.getNodesAndPorts();
		List<Object> nodeAndPortList;

		// retrieve size modifier
		final int dimHalf = scale;
		final int dim = dimHalf * 2;

		// loop variables //
		String[] nodeProps;
		KNode node;
		String parentId;
		KNode parent;
		KPort port;
		KShapeLayout layout;

		boolean isNodeFamily;
		int width;
		int height;
		int portCount;
		int portType;
		int portX;
		int portY;

		int i;
		int n;
		int p;
		int lp;
		final int l = nodeInfo.length;

		// nodeID : KNode, ParentID

		// -- 1. Define nodes by the properties, given in the string --
		for (n = 0; n < l; n++) {

			// i is our property array-pointer
			i = 0;
			nodeProps = nodeInfo[n].split(" ");

			// add entry to map
			nodeAndPortList = new ArrayList<Object>();
			nodesAndPorts.put(nodeProps[i++], nodeAndPortList);
			// create empty node
			node = KimlUtil.createInitializedNode();
			nodeAndPortList.add(node);

			// get parentID
			parentId = nodeProps[i++];
			if ("-1".equals(parentId)) {
				parentId = "";
				node.setParent(layoutInformation.getGraph());
			}
			nodeAndPortList.add(parentId);

			// get width and height
			width = Integer.parseInt(nodeProps[i++]);
			height = Integer.parseInt(nodeProps[i++]);

			// get Custom Node tag
			isNodeFamily = "f".equals(nodeProps[i++]);

			// set width and height of KNode
			layout = node.getData(KShapeLayout.class);
			layout.setWidth(width);
			layout.setHeight(height);

			// if node is a NodeFamily, go on and look for ports
			if (isNodeFamily) {

				// nodeFamily size is fixed
				layout.setProperty(LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_POS);
				portX = -dimHalf;

				// add Ports
				for (portType = 0; portType < 3; portType++) {
					// determine port x-position
					if (portType == 1) {
						portX = width - dimHalf;
					}

					portCount = Integer.parseInt(nodeProps[i++]);
					portY = Integer.parseInt(nodeProps[i++]);

					for (p = 0, lp = portCount; p < lp; p++) {

						// create port
						port = KimlUtil.createInitializedPort();
						port.setNode(node);
						nodeAndPortList.add(port);

						// set layout for Port
						layout = port.getData(KShapeLayout.class);

						// set position and dimension of port
						layout.setWidth(dim);
						layout.setHeight(dimHalf);
						layout.setYpos(portY);
						layout.setXpos(portX);

						portY += dim;
					}
				}
			}
		}

		// -- 2. resolve stored parentIDs to actual parents
		final Iterator<Entry<String, List<Object>>> it = nodesAndPorts.entrySet().iterator();
		while (it.hasNext()) {
			nodeAndPortList = (List<Object>) ((Entry) it.next()).getValue();

			// get KNode and its parentID
			node = (KNode) nodeAndPortList.get(0);
			parentId = nodeAndPortList.get(1).toString();

			// set other KNode as parent
			if (!parentId.isEmpty()) {
				parent = (KNode) nodesAndPorts.get(parentId).get(0);
				node.setParent(parent);
			}
		}
	}

	/**
	 * Parses an array which must contain quadruples of numbers.<br/>
	 * The first number must be the index of the children-array, where
	 * the source node is located.<br/>
	 * The second number must be the index of the children-array, where
	 * the target node is located.<br/>
	 * The third number must be the index of the source node List, where
	 * the source port is located.<br/>
	 * The fourth number must be the index of the target node List, where
	 * the target port is located.
	 * 
	 * @param edgeInfo
	 *            The single elements for the edges.
	 * @param layoutInformation
	 *            The necessary information about the graph.
	 * @throws UninitializedGraphException
	 *             If the graph has not yet been initialized.
	 */
	private static void addEdges(final String[] edgeInfo, final LayoutInformation layoutInformation) throws NumberFormatException {
		final Map<String, List<Object>> nodesAndPorts = layoutInformation.getNodesAndPorts();

		// set up loop variables
		String[] edgeProps;
		KEdge edge;

		String id;
		boolean hasSourcePort;
		Object[] edgeList;

		KNode sourceNode;
		int sourcePortIndex;
		KPort sourcePort = null;
		List<Object> sourceList;

		int targetPortIndex;
		KPort targetPort;
		List<Object> targetList;

		int ea = 0;
		int p;
		int epl;

		// read from String
		for (final String element : edgeInfo) {

			// get edge info of a single source
			edgeList = layoutInformation.getEdges()[ea++];
			edgeProps = element.split(" ");
			p = 0;

			// get source index and port
			id = edgeProps[p++];
			sourceList = nodesAndPorts.get(id);
			sourceNode = (KNode) sourceList.get(0);

			sourcePortIndex = Integer.parseInt(edgeProps[p++]);
			edgeList[2] = sourcePortIndex;

			hasSourcePort = ++sourcePortIndex != 0;

			if (hasSourcePort) {
				sourcePort = (KPort) sourceList.get(++sourcePortIndex);
			}

			// get target ports
			epl = edgeProps.length;
			while (p < epl) {

				// init edge
				edge = KimlUtil.createInitializedEdge();
				edgeList[0] = edge;
				edgeList[1] = id;

				// get target index and port
				id = edgeProps[p++];
				edgeList[3] = id;
				targetList = nodesAndPorts.get(id);

				// select connected nodes
				edge.setSource(sourceNode);
				edge.setTarget((KNode) targetList.get(0));

				targetPortIndex = Integer.parseInt(edgeProps[p++]);
				edgeList[4] = targetPortIndex;

				if (++targetPortIndex != 0) {
					targetPort = (KPort) targetList.get(++targetPortIndex);

					// set target port of edge
					edge.setTargetPort(targetPort);
					targetPort.getEdges().add(edge);
				}

				if (hasSourcePort) {
					// set source port of edge
					edge.setSourcePort(sourcePort);
					sourcePort.getEdges().add(edge);
				}

			}

		}
	}

	private static String[] correctEmptyArray(final String[] arr) {
		if ((arr.length == 1) && arr[0].isEmpty()) {
			return new String[0];
		} else {
			return arr;
		}
	}

	/**
	 * Reads the (layouted) positions of a constructed KGraph and writes them as integers to a string, seperating each number with a space.
	 * 
	 * @param layoutInformation
	 *            The object containing the necessary information about the graph.
	 * 
	 * @return The constructed string.
	 */
	private static String getPositions(final LayoutInformation layoutInformation) {
		final Map<String, List<Object>> nodesAndPorts = layoutInformation.getNodesAndPorts();
		final StringBuilder sb = new StringBuilder();

		// Run through all available entries
		for (final Map.Entry<String, List<Object>> entry : nodesAndPorts.entrySet()) {
			final String id = entry.getKey();
			final List<Object> nodeAndPortList = entry.getValue();

			final KShapeLayout layout = ((KNode) nodeAndPortList.get(0)).getData(KShapeLayout.class);

			// Get the position of the node
			final int posX = Math.round(layout.getXpos() + (layout.getWidth() / 2));
			final int posY;
			// HotFix: wrong nodeFamily position: needs to be half the height lower
			if (nodeAndPortList.size() > 2) {
				posY = Math.round(layout.getYpos() + (layout.getHeight()));
			} else {
				posY = Math.round(layout.getYpos() + (layout.getHeight() / 2));
			}

			// Get the size of the node
			final int width = Math.round(layout.getWidth());
			final int height = Math.round(layout.getHeight());

			// Put everything into the string
			sb.append(id);
			sb.append(" ");
			sb.append(posX);
			sb.append(" ");
			sb.append(posY);
			sb.append(" ");
			sb.append(width);
			sb.append(" ");
			sb.append(height);
			sb.append(";");
		}

		return sb.toString();
	}

	/**
	 * This method iterates through all edges of the given object and creates a semicolon separated String containing bend points of all edges.
	 * 
	 * @param layoutInformation
	 *            The object containing the information about the graph and the layout.
	 * @return The string containing the bend points.
	 */
	private static String getBendPoints(final LayoutInformation layoutInformation) {
		final StringBuilder sb = new StringBuilder();
		final Object[][] edges = layoutInformation.getEdges();

		KEdge edge;
		int e;

		final int l = edges.length;
		final int el = 4;

		// iterate edges
		for (int i = 0; i < l; i++) {
			e = 0;
			edge = (KEdge) edges[i][e];

			final KEdgeLayout layout = edge.getData(KEdgeLayout.class);
			final EList<KPoint> bendPoints = layout.getBendPoints();

			// add IDs and portIndices
			while (++e < el) {
				sb.append(edges[i][e].toString());
				sb.append(" ");
			}
			sb.append(edges[i][4].toString());

			// iterate bend points
			final int bl = bendPoints.size();
			KPoint point;

			if (bl > 0) {

				point = bendPoints.get(0);
				sb.append(" ");
				sb.append(Math.round(point.getX()));
				sb.append(" ");
				sb.append(Math.round(point.getY()));

				for (int b = 1; b < bl; b++) {
					sb.append(" ");
					point = bendPoints.get(b);
					sb.append(Math.round(point.getX()));
					sb.append(" ");
					sb.append(Math.round(point.getY()));

				}
			}
			sb.append(";");
		}

		return sb.toString();
	}

	/**
	 * A container for the layout information which will be assembled during the layout process.
	 * 
	 * @author Nils Christian Ehmke
	 */
	private static class LayoutInformation {

		private final KNode graph;
		/**
		 * An array containing the child node and the ports from left to right.
		 */
		private final Map<String, List<Object>> children;
		/**
		 * An array containing edges. This way we can return edge information in the same order.
		 */
		private final Object[][] edges;

		/**
		 * Creates a new instance of this class.
		 * 
		 * @param graph
		 *            The graph itself.
		 * @param children
		 *            The child notes and the ports.
		 * @param edges
		 *            The edges of the graph.
		 */
		public LayoutInformation(final KNode graph, final Map<String, List<Object>> children, final Object[][] edges) { // NOPMD (array)
			this.graph = graph;
			this.children = children;
			this.edges = edges;

			// set graph layout options
			final KShapeLayout layout = graph.getData(KShapeLayout.class);
			// layout.setProperty(LayoutOptions.LAYOUT_HIERARCHY, true);
			layout.setProperty(LayoutOptions.EDGE_ROUTING, EdgeRouting.ORTHOGONAL);

		}

		/**
		 * Getter for the property {@link LayoutInformation#graph}.
		 * 
		 * @return The current value of the property.
		 */
		public KNode getGraph() {
			return this.graph;
		}

		/**
		 * Getter for the property {@link LayoutInformation#children}.
		 * 
		 * @return The current value of the property.
		 */
		public Map<String, List<Object>> getNodesAndPorts() {
			return this.children;
		}

		/**
		 * Getter for the property {@link LayoutInformation#edges}.
		 * 
		 * @return The current value of the property.
		 */
		public Object[][] getEdges() {
			return this.edges; // NOPMD (array)
		}

	}
}
